#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdint.h>
#include <errno.h>
#include <linux/kdbus.h>

#define KDBUS_SYSTEM_BUS_PATH "/sys/fs/kdbus/0-system/bus"

#define KDBUS_ALIGN8(l) (((l) + 7) & ~7)
#define KDBUS_ITEM_HEADER_SIZE offsetof(struct kdbus_item, data)
#define KDBUS_ITEM_SIZE(s) KDBUS_ALIGN8((s) + KDBUS_ITEM_HEADER_SIZE)
#define KDBUS_ITEM_NEXT(item) \
	(typeof(item))(((uint8_t *)item) + KDBUS_ALIGN8((item)->size))
#define KDBUS_FOREACH(iter, first, _size)				\
	for (iter = (first);						\
	     ((uint8_t *)(iter) < (uint8_t *)(first) + (_size)) &&	\
	       ((uint8_t *)(iter) >= (uint8_t *)(first));		\
	     iter = (void*)(((uint8_t *)iter) + KDBUS_ALIGN8((iter)->size)))

#define POOL_SIZE (16 * 1024 * 1024)

#define err_r(_r, _msg) fprintf(stderr, "log: %d, %s, %s, %d, %s\n", (_r), (_msg), __func__, __LINE__, __FILE__)
#define err(_msg) err_r(errno, (_msg))

static uint8_t* pool;
static char* server_names[16];
int ready_fd = 0;

static inline int kdbus_cmd_hello(int bus_fd, struct kdbus_cmd_hello *cmd)
{
	int ret = ioctl(bus_fd, KDBUS_CMD_HELLO, cmd);
	return (ret < 0) ? (errno > 0 ? -errno : -EINVAL) : 0;
}

static inline int kdbus_cmd_free(int conn_fd, struct kdbus_cmd_free *cmd)
{
	int ret = ioctl(conn_fd, KDBUS_CMD_FREE, cmd);
	return (ret < 0) ? (errno > 0 ? -errno : -EINVAL) : 0;
}

static inline int kdbus_cmd_name_acquire(int conn_fd, struct kdbus_cmd *cmd)
{
	int ret = ioctl(conn_fd, KDBUS_CMD_NAME_ACQUIRE, cmd);
	return (ret < 0) ? (errno > 0 ? -errno : -EINVAL) : 0;
}

static void bus_poool_free_slice(int fd, uint64_t offset)
{
	struct kdbus_cmd_free cmd = {
		.size = sizeof(cmd),
		.offset = offset,
	};
	int r;
	r = kdbus_cmd_free(fd, &cmd);
	if (r < 0)
		err_r(r, "cannot free pool slice");
}

static int bus_open_connection(const char *name,
			       uint64_t recv_flags)
{
	struct kdbus_cmd_hello hello;
	int r;
	int fd;

	fd = open(name, O_RDWR | O_CLOEXEC);
	if (fd < 0) {
		err("cannot open bus");
		r = -errno;
		goto error;
	}

	memset(&hello, 0, sizeof(hello));
	hello.size = sizeof(hello);
	hello.attach_flags_send = _KDBUS_ATTACH_ALL;
	hello.attach_flags_recv = recv_flags;
	hello.pool_size = POOL_SIZE;
	r = kdbus_cmd_hello(fd, &hello);
	if (r < 0) {
		err_r(r, "HELLO failed");
		close(fd);
		goto error;
	}

	bus_poool_free_slice(fd, hello.offset);

	/*
	 * Map the pool of the connection. Its size has been set in the
	 * command struct above. See kdbus.pool(7).
	 */
	pool = mmap(NULL, POOL_SIZE, PROT_READ, MAP_SHARED, fd, 0);
	if (pool == MAP_FAILED) {
		err("cannot mmap pool");
		r = -errno;
		close(fd);
		goto error;
	}

	return fd;

error:
	return r;
}



static int bus_acquire_name(int fd, const char *name)
{
	struct kdbus_item *item;
	struct kdbus_cmd *cmd;
	size_t size;
	int r;

	/*
	 * This function acquires a well-known name on the bus through the
	 * KDBUS_CMD_NAME_ACQUIRE ioctl. This ioctl takes an argument of type
	 * 'struct kdbus_cmd', which is assembled below. See kdbus.name(7).
	 */
	size = sizeof(*cmd);
	size += KDBUS_ITEM_SIZE(strlen(name) + 1);

	cmd = alloca(size);
	memset(cmd, 0, size);
	cmd->size = size;

	/*
	 * The command requires an item of type KDBUS_ITEM_NAME, and its
	 * content must be a valid bus name.
	 */
	item = cmd->items;
	item->type = KDBUS_ITEM_NAME;
	item->size = KDBUS_ITEM_HEADER_SIZE + strlen(name) + 1;
	strncpy(item->str, name, strlen(name) + 1);

	/*
	 * Employ the command on the connection owner file descriptor.
	 */
	r = kdbus_cmd_name_acquire(fd, cmd);
	if (r < 0)
		return err_r(r, "cannot acquire name");

	return 0;
}


int parse_args(int argc, char** argv)
{
	char state = 0;
	unsigned long int uid = 0;
	unsigned long int gid = 0;
	int i = 1;
	int r = 0;
	int fdl = -1;
	int count = 0;

	while (i < argc) {
		if (argv[i][0] == '-')
			state = argv[i][1];
		else {
			switch (state) {

			case 'n':
				server_names[count++] = argv[i];
				break;

			case 'u':
				uid = strtoul(argv[i], NULL, 10);
				r = setuid((uid_t)uid);
				if (r < 0) {
					fprintf(stderr, "Cannot set uid");
					exit(-3245);
				}
			break;

			case 'g':
				gid = strtoul(argv[i], NULL, 10);
				r = setgid((gid_t)gid);
				if (r < 0) {
					fprintf(stderr, "Cannot set gid");
					exit(-3245);
				}
				break;

			case 'l':
				fdl = open("/proc/self/attr/current", 0, S_IWUSR);
				if (fdl < 0) {
					fprintf(stderr, "Cannot open /proc/self/attr/current\n");
					exit(-345);
				}

				r = write(fdl, argv[i], strlen(argv[i]));
				if (r < 0) {
					fprintf(stderr, "Cannot write to /proc/self/attr/current\n");
					close(fdl);
					exit(-345);
				}
				close(fdl);
			break;
			case 'q':
				ready_fd = strtol(argv[i], NULL, 10);
				break;
			}
		}
		i++;
	}
	return 0;
}

int main(int argc, char* argv[])
{
	int j = 0;
	int fd;
	parse_args(argc, argv);
	fd = bus_open_connection(KDBUS_SYSTEM_BUS_PATH, KDBUS_ATTACH_PIDS);
	if (fd < 0)
		return -1;

	while (server_names[j] != '\0') {
		if (bus_acquire_name(fd, server_names[j]))
			return -1;
		j++;
	}

	close(ready_fd);

	for (;;);
	return 0;
}
