/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#include "cynara.hpp"
#include "groups_proxy.hpp"
#include "naive_policy_checker.hpp"
#include "serializer.hpp"
#include "tslog.hpp"

/**
 * \file
 * \ingroup Implementation
 */


DEF_NODESTRUCT_GLOBAL(ldp_xml_parser::NaivePolicyChecker, policy_checker_system)
DEF_NODESTRUCT_GLOBAL(ldp_xml_parser::NaivePolicyChecker, policy_checker_session)

namespace ldp_xml_parser {

template <typename T>
VGid &NaivePolicyChecker::getMapGroup(uid_t uid)
{
	return mapGroups[type_list::IDX<T, MatchItemTypes>::idx][uid];
}

void NaivePolicyChecker::updateSupplementaryGroups(const VGid &groups, uid_t uid, gid_t gid) {
	auto &vsend = getMapGroup<MatchItemSend>(uid);
	auto &vrecv = getMapGroup<MatchItemReceive>(uid);
	auto &vaccess = getMapGroup<MatchItemAccess>(uid);

	if (groups.empty()) {
		vsend.push_back(gid);
		vrecv.push_back(gid);
		vaccess.push_back(gid);
		return;
	}

	/* insert supplementary group */
	for (const auto group : groups) {
		if (m_bus_db.existsPolicyForGroup<MatchItemSend>(group))
			vsend.push_back(group);
		if (m_bus_db.existsPolicyForGroup<MatchItemReceive>(group))
			vrecv.push_back(group);
		vaccess.push_back(group);	// no filtering, it will be used once
	}

	if (vsend.empty())
		vsend.push_back(-1);
	if (vrecv.empty())
		vrecv.push_back(-1);
}

void NaivePolicyChecker::updateSupplementaryGroupsOwn(const VGid &groups, uid_t uid, gid_t gid) {
	auto &vown = getMapGroup<MatchItemOwn>(uid);
	if (groups.empty()) {
		vown.push_back(gid);
	} else {
		for (const auto group : groups) {
			if (m_bus_db.existsPolicyForGroup<MatchItemOwn>(group))
				vown.push_back(group);
		}
	}
}

template <>
const VGid *NaivePolicyChecker::getGroups<MatchItemOwn>(uid_t uid, gid_t)
{
	return &getMapGroup<MatchItemOwn>(uid);
}

template <>
const VGid *NaivePolicyChecker::getGroups<MatchItemAccess>(uid_t uid, gid_t)
{
	return &getMapGroup<MatchItemAccess>(uid);
}

template <typename T>
const VGid *NaivePolicyChecker::getGroups(uid_t uid, gid_t gid)
{
	if (uid == getuid() && gid == getgid())
		return &getMapGroup<T>(uid);

	pthread_mutex_lock(&mutexGroup);
	auto &vgid = getMapGroup<T>(uid);

	if (vgid.empty())
		updateSupplementaryGroups(get_groups(uid, gid), uid, gid);
	pthread_mutex_unlock(&mutexGroup);

	if (vgid[0] == (gid_t)-1)
		return nullptr;

	return &vgid;
}

DecisionResult NaivePolicyChecker::parseDecision(const DecisionItem& decision,
									   uid_t uid,
									   const char* label) const {
	tslog::log_verbose("----Decision made\n");

	switch (decision.getDecision())
	{
		case Decision::ALLOW:
			return DecisionResult::ALLOW;
		case Decision::CHECK:
			switch (ldp_cynara::Cynara::check(label, decision.getPrivilege(), uid))
			{
				case ldp_cynara::CynaraResult::ALLOW:
					return DecisionResult::ALLOW;
				case ldp_cynara::CynaraResult::DENY:
					return DecisionResult::DENY;
				default:
					return DecisionResult::CYNARA_ERROR;
			}
		default:
			break;
	}
	return DecisionResult::DENY;
}

DecisionItem NaivePolicyChecker::checkItemAccess(const MatchItemAccess& item)
{
	DecisionItem ret = m_bus_db.getDecisionItemContextMandatory(item);
	// access rules can be defined only in default/mandatory context
	// defining them elsewhere is considered as policy syntax error by dbus-daemon
	// thus, no checking in user or group policies
	if (ret.getDecision() == Decision::ANY)
		ret = m_bus_db.getDecisionItemContextDefault(item);

	return ret;
}

DecisionResult NaivePolicyChecker::check(uid_t bus_owner,
							   uid_t uid,
							   gid_t gid,
							   const char* const label) {
	const auto &gids = *getGroups<MatchItemAccess>(uid, gid);
	auto ret = checkItemAccess(MatchItemAccess(uid, gids));
	if (ret.getDecision() == Decision::ANY) {
		if (bus_owner == uid) {
			ret = Decision::ALLOW;
		}
	}
	return parseDecision(ret, uid, label);
}

template <typename T>
DecisionResult NaivePolicyChecker::check(uid_t uid,
							   gid_t gid,
							   const char* const label,
							   const T &matchItem) {
	auto ret = checkItem(uid, gid, matchItem);
	return parseDecision(ret, uid, label);
}
template DecisionResult NaivePolicyChecker::check(uid_t, gid_t, const char *, const MatchItemOwn &);
template DecisionResult NaivePolicyChecker::check(uid_t, gid_t, const char *, const MatchItemSend &);
template DecisionResult NaivePolicyChecker::check(uid_t, gid_t, const char *, const MatchItemReceive &);

template<typename T>
DecisionItem NaivePolicyChecker::checkItem(uid_t uid, gid_t gid, const T& item) {
	DecisionItem ret = m_bus_db.getDecisionItemContextMandatory(item);

	if (ret.getDecision() == Decision::ANY)
		ret = m_bus_db.getDecisionItemUser(uid, item);

	if (ret.getDecision() == Decision::ANY)
		ret = checkGroupPolicies<T>(uid, gid, item);

	if (ret.getDecision() == Decision::ANY)
		ret = m_bus_db.getDecisionItemContextDefault(item);

	return ret;
}

template<typename T>
DecisionItem NaivePolicyChecker::checkGroupPolicies(uid_t uid, gid_t gid, const T& item) {
	const auto *sgroups = getGroups<T>(uid, gid);
	if (sgroups == nullptr)
		return Decision::ANY;

	for (const auto sgid : *sgroups) {
		DecisionItem ret = m_bus_db.getDecisionItemGroup(sgid, item);

		if (ret.getDecision() != Decision::ANY)
			return ret;
	}
	return Decision::ANY;
}

void NaivePolicyChecker::updateGroupDb(uid_t uid, gid_t gid)
{
	pthread_mutex_lock(&mutexGroup);

	for (auto &mapGroup : mapGroups)
		mapGroup.clear();

	auto groups = get_groups(uid, gid);
	updateSupplementaryGroups(groups, uid, gid);
	updateSupplementaryGroupsOwn(groups, uid, gid);

	pthread_mutex_unlock(&mutexGroup);
}

bool NaivePolicyChecker::initDb(const char *config_name, const char *serialized_filename)
{
	assert(config_name);
	assert(config_name[0]);

	bool result = false;
	m_bus_db.release();

	if (serialized_filename) {
		if (m_bus_db.init(serialized_filename))
			result = true;
	}

	if (!result) {
		tslog::logWarning(std::string(serialized_filename?:"<no_serialized_filename>")
				.append(": serialized policy file not found, using XML policy file"));

		// fallback - we have only XML files
		result = m_bus_db.initFromXML(config_name);
	}

	if (result && tslog::enabled()) {
		printContent();
	}

	return result;
}

void NaivePolicyChecker::printContent()
{
	m_bus_db.printContent();

	for (const auto &mapGroup : mapGroups) {
		std::cerr << std::endl << "---- mapGroup ----" << std::endl;
		int size = sizeof(mapGroup);
		size += mapGroup.size() * sizeof(typename std::remove_reference<decltype(mapGroup)>::type::value_type);

		for (const auto& i : mapGroup) {
			size += i.second.capacity() * sizeof(gid_t);
			std::cerr << "uid: " << i.first << " | gids: ";
			for (auto j : i.second)
				std::cerr << " " << j << ",";
			std::cerr << std::endl;
		}
		std::cerr << "Memory consumption: " << size << " B" << std::endl;
	}
}

} // namespace ldp_xml_parser
