/*
 * Copyright (c) 2018-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#include "own_tree.hpp"
#include <boost/tokenizer.hpp>
#include <deque>
#include <iostream>
#include <unordered_map>

/**
 * \file
 * \ingroup Implementation
 */

using namespace ldp_xml_parser;




OwnershipTree::OwnershipTree() {
	__root = std::make_shared<TreeNode>("");
}

std::deque<std::string> tokenize(const std::string& item_name)
{
	std::deque<std::string> elements;
	typedef boost::tokenizer<boost::char_separator<char>> tokenizer;
	boost::char_separator<char> separator(".");
	tokenizer tokens(item_name, separator);
	for (auto token: tokens) {
		elements.push_back(token);
	}
	return elements;
}

void OwnershipTree::addItem(const ItemOwn &item) {
	if (item.isMatchAll()) {
		if (item.getDecision().getDecision() != Decision::ANY){
			__root->setDecisionItem(item.getDecision(), true);
		}
		return;
	}

	auto tokens = tokenize(item.getName());
	__root->add(tokens, item.getDecision(), item.isPrefix());
}

TreeNode::TreeNode(const std::string &token)
:__token(token), __own_prefix_decision_item(Decision::ANY), __own_decision_item(Decision::ANY)
{}

void TreeNode::add(std::deque<std::string>& tokens, const DecisionItem& decision, const bool is_prefix)
{
	if (tokens.empty()) {
		setDecisionItem(decision, is_prefix);

		if (is_prefix) {
			__children.clear();
		}
		return;
	}

	const auto& token = tokens.front();
	auto it = __children.find(token);
	if (it == __children.end()) {
		it = __children.insert({token, std::make_shared<TreeNode>(token)}).first;
	}
	assert(it != __children.end());
	tokens.pop_front();

	it->second->add(tokens, decision, is_prefix);
}

void TreeNode::setDecisionItem(const DecisionItem& decision, const bool is_prefix)
{
	if (is_prefix) {
		__own_prefix_decision_item = decision;
		__own_decision_item = Decision::ANY;
	} else {
		__own_decision_item = decision;
		// we don't clear own_prefix decision, it still applies to children
	}
}

const std::string &TreeNode::getToken() const {
	return __token;
}

const DecisionItem &TreeNode::getOwnPrefixDecisionItem() const {
	return __own_prefix_decision_item;
}

const DecisionItem &TreeNode::getOwnDecisionItem() const {
	return __own_decision_item;
}

const std::unordered_map<std::string, std::shared_ptr<TreeNode>> &TreeNode::getChildren() const {
	return __children;
}
