/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
#include "include/fb_generated.h"
#include "print_content.hpp"
#include "serialized_convert.hpp"
#include "serializer.hpp"
#include "storage_backend_serialized.hpp"
#include "transaction_guard.hpp"
#include "tslog.hpp"
#include <boost/tokenizer.hpp>
#include <cerrno>
#include <fcntl.h>
#include <map>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

using namespace FB;
using ldp_xml_parser::MatchItemSend;

namespace ldp_serialized {

// Set max size of serialized file to prevent mmap with unexpected memory size.
// 1MB. Adjustable
#define MAX_SFILE_SIZE (1024 * 1024)

typedef boost::tokenizer<boost::char_separator<char>> tokenizer;

template <typename T>
struct type_helper;

class StorageBackendSerialized::StorageBackendSerializedImpl {
	int fd{-1};
	uint8_t *mem{static_cast<decltype(mem)>(MAP_FAILED)};
	size_t length{0};
	const FB::File *file{nullptr};

	std::unique_ptr<ldp_xml_parser::Serializer> serializer;

	void releaseMMap();
	void releaseFD();

	typedef std::pair<const ItemSend *, unsigned> ItemWithScore;	// a pair: Item, index of Item
	typedef std::vector<ItemWithScore> ItemsWithScore;				// items
	typedef std::pair<unsigned, ItemsWithScore> HighestScoreWithItems;	// value for index map: highest index for items, items
	typedef std::map<boost::string_ref, HighestScoreWithItems> MapIndex;
	typedef std::vector<ItemWithScore> SendPrefixIndex;

	MapIndex sendIndex; // context default
	SendPrefixIndex sendPrefixIndex; // context default prefix rules
	std::map<uid_t, MapIndex> userSendIndex;
	std::map<uid_t, SendPrefixIndex> userSendPrefixIndex;

public:
	bool init(const char *filename, bool verify);
	bool init(const FB::File *f);
	bool initFromXML(const char *config_name);
	void release();

	void createSendIndex();
	ldp_xml_parser::DecisionItem getDecisionFromSendIndex(const MatchItemSend &item, const MapIndex &index, const SendPrefixIndex &prefixIndex);
	ldp_xml_parser::DecisionItem getDecisionFromSendIndex(const MatchItemSend &item);
	ldp_xml_parser::DecisionItem getDecisionFromSendIndex(const MatchItemSend &item, uid_t uid);

	void printContent(const bool xml_format = false) const;

	template <typename T, typename M = typename type_helper<T>::policy_set_type>
	const M *getPolicySet();
};

inline boost::string_ref s(const flatbuffers::String *str) {
	return boost::string_ref(str->c_str(), str->size());
}

template <typename T, typename I>
bool match(const T &match, const I *i) {
	return match.match(makeMessageType(i->type()),
					   s(i->interface()),
					   s(i->path()),
					   s(i->member()),
					   s(i->name()),
					   i->is_name_prefix(),
					   makeDecision(i->decision()->decision()));
}

ldp_xml_parser::DecisionItem StorageBackendSerialized::StorageBackendSerializedImpl::getDecisionFromSendIndex(const MatchItemSend &item,
		const MapIndex &index, const SendPrefixIndex &prefixIndex) {
	ldp_xml_parser::DecisionItem decision(ldp_xml_parser::Decision::ANY);

	ItemWithScore currentBest(nullptr, 0);

	auto updateCurrentBest = [&currentBest, &item, &index](boost::string_ref name) {
		// check if there are any rules for the name
		auto fit = index.find(name);
		if (fit == index.end())
			return;

		// check if there's any chance to get better score
		// fit->second.first is the highest score from vector fit->second.second
		if (fit->second.first <= currentBest.second)
			return;

		// look for better score
		for (const auto &it: fit->second.second) {
			if (it.second > currentBest.second && match(item, it.first))
				currentBest = it;
		}
	};

	// iterate over names
	for (const auto &name: item.getNames()) {
		// find and check the no-prefix rules
		updateCurrentBest(name);

		// check the prefix rules
		for (const auto &prefixItem: prefixIndex) {
			if (prefixItem.second > currentBest.second && match(item, prefixItem.first))
				currentBest = prefixItem;
		}
	}

	// check no-name rules
	updateCurrentBest("");

	if (currentBest.first)
		decision = makeDecisionItem(currentBest.first->decision());

	return decision;
}

ldp_xml_parser::DecisionItem StorageBackendSerialized::StorageBackendSerializedImpl::getDecisionFromSendIndex(const MatchItemSend &item) {
	return getDecisionFromSendIndex(item, sendIndex, sendPrefixIndex);
}

ldp_xml_parser::DecisionItem StorageBackendSerialized::StorageBackendSerializedImpl::getDecisionFromSendIndex(const MatchItemSend &item, uid_t uid) {
	return getDecisionFromSendIndex(item, userSendIndex[uid], userSendPrefixIndex[uid]);
}

void StorageBackendSerialized::StorageBackendSerializedImpl::createSendIndex() {
	// helper for adding items to indexes
	auto add = [](MapIndex &index, SendPrefixIndex &prefixIndex, boost::string_ref key, ItemWithScore value) {
		if (value.first->is_name_prefix()) {
			prefixIndex.push_back(value);
		} else {
			index[key].second.push_back(value);
			// we insert items in increasing score order, so we just update the highest score on each add
			index[key].first = value.second;
		}
	};

	// context default index
	const auto *policy = file->m_send_set()->context_default();
	unsigned cnt = 1;

	const auto *v = policy->items();
	for (auto it = v->begin(); it != v->end(); ++it)
		add(sendIndex, sendPrefixIndex,
				boost::string_ref(it->name()->c_str(), it->name()->size()),
				ItemWithScore(*it, cnt++));

	// users index
	for (const auto &uidPolicy: *file->m_send_set()->user()) {
		cnt = 1;
		for (const auto it: *uidPolicy->policy()->items()) {
			add(userSendIndex[uidPolicy->id()],
					userSendPrefixIndex[uidPolicy->id()],
					boost::string_ref(it->name()->c_str(), it->name()->size()),
					ItemWithScore(it, cnt++));
		}
	}
}

void StorageBackendSerialized::StorageBackendSerializedImpl::releaseMMap() {
	assert(MAP_FAILED != mem);
	assert(0 != length);

	if (munmap(mem, length) != 0)
		tslog::log_error("munmap(): ", tslog::print_errno(errno), "\n");

	mem = static_cast<decltype(mem)>(MAP_FAILED);
	length = 0;
}

void StorageBackendSerialized::StorageBackendSerializedImpl::releaseFD() {
	assert(-1 != fd);

	if (close(fd) != 0)
		tslog::log_error("close(): ", tslog::print_errno(errno), "\n");

	fd = -1;
}

void StorageBackendSerialized::StorageBackendSerializedImpl::release() {
	if (-1 != fd) {		// we need to check it, because we may have initialized the storage directly from File *
		releaseMMap();
		releaseFD();
	}

	if (nullptr != serializer.get()) {
		serializer.reset(nullptr);
	}

	file = nullptr;
}

bool StorageBackendSerialized::StorageBackendSerializedImpl::init(const char *filename, bool verify) {
	assert(nullptr == file);

	auto err = [filename] (const char *what) {
		tslog::log_error("Can't ", what, " ", filename, ": ", tslog::print_errno(errno), "\n");
		return false;
	};

	fd = open(filename, O_RDONLY);
	if (-1 == fd)
		return err("open");

	auto openGuard = transaction_guard::makeGuard([&] () { releaseFD(); });

	struct stat buf;

	if (fstat(fd, &buf) == -1)
		return err("stat");

	length = buf.st_size;
	if (length > MAX_SFILE_SIZE) {
		tslog::log_error("Serialized file size(", length, ") is too large. (>", MAX_SFILE_SIZE, ") bytes.\n");
		return false;
	}

	mem = reinterpret_cast<uint8_t*>(mmap(NULL, length, PROT_READ, MAP_PRIVATE, fd, 0));
	if (MAP_FAILED == mem)
		return err("mmap");

	auto mmapGuard = transaction_guard::makeGuard([&] () { releaseMMap(); });

	if (verify) {
		auto verifier = flatbuffers::Verifier(mem, length);
		if (!FB::VerifyFileBuffer(verifier) || !FB::FileBufferHasIdentifier(mem)) {
			tslog::log_error("verification of serialized data: failed\n");
			return false;
		}
	}

	openGuard.dismiss();
	mmapGuard.dismiss();

	file = GetFile(mem);
	bool res = file != nullptr;
	if (res)
		createSendIndex();
	return res;
}

bool StorageBackendSerialized::StorageBackendSerializedImpl::init(const FB::File *f) {
	assert(nullptr == file);

	file = f;
	createSendIndex();
	return true;
}

bool StorageBackendSerialized::StorageBackendSerializedImpl::initFromXML(const char *config_name) {
	assert(nullptr == file);
	assert(nullptr == serializer.get());

	serializer.reset(new ldp_xml_parser::Serializer());

	size_t serialized_size;
	uint8_t *data = serializer->serialize(config_name, serialized_size);
	if (nullptr == data)
		return false;

	return init(FB::GetFile(data));
}

void StorageBackendSerialized::StorageBackendSerializedImpl::printContent(const bool xml_format) const {
	print_content::use_xml_format(xml_format);
	std::cerr << *file;
}

/*************************************************/

#define TYPE_HELPER(T, t) \
	template <> \
	struct type_helper<ldp_xml_parser::MatchItem##T> { \
		typedef FB::T##Set policy_set_type; \
		typedef FB::Policy##T policy_type; \
	}; \
	template <> \
	auto StorageBackendSerialized::StorageBackendSerializedImpl::getPolicySet<ldp_xml_parser::MatchItem##T>() \
		-> const typename type_helper<ldp_xml_parser::MatchItem##T>::policy_set_type * { \
		assert(file); \
		return file->m_##t##_set(); \
	}

TYPE_HELPER(Own, own)
TYPE_HELPER(Send, send)
TYPE_HELPER(Receive, receive)
TYPE_HELPER(Access, access)

bool StorageBackendSerialized::init(const char *filename, bool verify) {
	return pimpl->init(filename, verify);
}

bool StorageBackendSerialized::init(const FB::File *f) {
	return pimpl->init(f);
}

bool StorageBackendSerialized::initFromXML(const char *config_name) {
	return pimpl->initFromXML(config_name);
}

void StorageBackendSerialized::release() {
	pimpl->release();
}

void StorageBackendSerialized::printContent(const bool xml_format) const {
	pimpl->printContent(xml_format);
}

template <> bool match(const ldp_xml_parser::MatchItemAccess &match, const FB::ItemAccess *item) {
	return match.match(makeBusAccessType(item->type()), item->uid(), item->gid());
}

template <typename T, typename P = typename type_helper<T>::policy_type>
ldp_xml_parser::DecisionItem getDecisionItem(const T &item, const P *policy) {
	const auto *v = policy->items();
	for (auto it = v->rbegin(); it != v->rend(); ++it) {
		if (match(item, *it))
			return makeDecisionItem((*it)->decision());
	}
	return ldp_xml_parser::Decision::ANY;
}

const DecisionItem *getDecisionItemFromTree(const FB::PolicyOwnNode *node,
											tokenizer &tokens,
											tokenizer::iterator &iterator) {
	if (iterator == tokens.end()) {
		if (node->decision_item()->decision() != Decision_ANY)
			return node->decision_item();
		else
			return node->prefix_decision_item();
	}

	auto child = node->children()->LookupByKey(iterator->c_str());
	if (nullptr == child)
		return node->prefix_decision_item();

	++iterator;
	const DecisionItem *child_decision = getDecisionItemFromTree(child, tokens, iterator);
	if (child_decision->decision() == Decision_ANY)
		return node->prefix_decision_item();

	return child_decision;
}

template <> ldp_xml_parser::DecisionItem getDecisionItem(const ldp_xml_parser::MatchItemOwn &item,
														 const PolicyOwn *policy) {
	if (item.getName().length() == 0)
		return ldp_xml_parser::Decision::DENY;

	boost::char_separator<char> separator(".");
	tokenizer tokens(item.getName(), separator);

	auto node = policy->tree();
	if (nullptr == node)
		return ldp_xml_parser::Decision::ANY;

	auto iterator = tokens.begin();

	return makeDecisionItem(getDecisionItemFromTree(node, tokens, iterator));
}

template <typename T, typename P = typename type_helper<T>::policy_type>
ldp_xml_parser::DecisionItem getDecisionItemMaybeNull(const T &item, const P *policyPair) {
	if (nullptr == policyPair)
		return ldp_xml_parser::Decision::ANY;
	return getDecisionItem(item, policyPair->policy());
}

template <typename T>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemContextMandatory(const T &item) const {
	return getDecisionItem(item, pimpl->getPolicySet<T>()->context_mandatory());
}

template <>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemContextDefault(const MatchItemSend &item) const {
	return pimpl->getDecisionFromSendIndex(item);
}

template <>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemUser(uid_t uid, const MatchItemSend &item) const {
	auto *policyPair = pimpl->getPolicySet<MatchItemSend>()->user()->LookupByKey(uid);
	if (nullptr == policyPair)
		return ldp_xml_parser::Decision::ANY;
	return pimpl->getDecisionFromSendIndex(item, uid);
}

template <typename T>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemContextDefault(const T &item) const {
	return getDecisionItem(item, pimpl->getPolicySet<T>()->context_default());
}

template <typename T>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemUser(uid_t uid, const T &item) const {
	return getDecisionItemMaybeNull(item, pimpl->getPolicySet<T>()->user()->LookupByKey(uid));
}

template <typename T>
ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemGroup(gid_t gid, const T &item) const {
	return getDecisionItemMaybeNull(item, pimpl->getPolicySet<T>()->group()->LookupByKey(gid));
}

template <typename T>
bool StorageBackendSerialized::existsPolicyForGroup(gid_t gid) const {
	return pimpl->getPolicySet<T>()->group()->LookupByKey(gid) != nullptr;
}

#define T_INSTANTIATION(T) \
	template ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemContextMandatory(const ldp_xml_parser::MatchItem##T &item) const; \
	template ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemContextDefault(const ldp_xml_parser::MatchItem##T &item) const;

T_INSTANTIATION(Own)
T_INSTANTIATION(Send)
T_INSTANTIATION(Receive)
T_INSTANTIATION(Access)

#undef T_INSTANTIATION

#define T_INSTANTIATION2(T) \
	template ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemUser(uid_t uid, const ldp_xml_parser::MatchItem##T &item) const; \
	template ldp_xml_parser::DecisionItem StorageBackendSerialized::getDecisionItemGroup(gid_t gid, const ldp_xml_parser::MatchItem##T &item) const; \
	template bool StorageBackendSerialized::existsPolicyForGroup<ldp_xml_parser::MatchItem##T>(gid_t) const;

T_INSTANTIATION2(Own)
T_INSTANTIATION2(Send)
T_INSTANTIATION2(Receive)

#undef T_INSTANTIATION2

StorageBackendSerialized::StorageBackendSerialized()
	: pimpl{new StorageBackendSerializedImpl} {
}

StorageBackendSerialized::~StorageBackendSerialized() {
	pimpl->release();
}

}
