/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/**
 * \file
 * \ingroup Implementation
 */
#include "tslog.hpp"
#include <cstring>

namespace tslog {

ldp_log_level g_verbosity{ldp_log_level::SILENT};
pthread_mutex_t g_mutex = PTHREAD_MUTEX_INITIALIZER;

bool get_log_env(char const *name) {
	char const *ldp_log_mode = getenv(name);
	return ldp_log_mode && '0' != *ldp_log_mode;
}

void init(ldp_log_level level) {
	g_verbosity = get_log_env("LDP_LOG") ? (get_log_env("LDP_VERBOSE") ? ldp_log_level::DEBUG : ldp_log_level::VERBOSE): level;
}

bool enabled(ldp_log_level level) { return level <= g_verbosity; }

void flush(ldp_log_level level) {
	if (tslog::enabled(level)) {
		pthread_mutex_lock(&g_mutex);
		std::cout << std::flush;
		pthread_mutex_unlock(&g_mutex);
	}
}

void logError(const std::string &error)
{
	log_error(error, "\n");
	LOGE("%s", error.c_str());
}

void logWarning(const std::string &warning)
{
	log_warning(warning, "\n");
	LOGW("%s", warning.c_str());
}

void logDebug(const std::string &debug)
{
	log_debug(debug, "\n");
	LOGD("%s", debug.c_str());
}

void logVerbose(const std::string &verbose)
{
	log_verbose(verbose, "\n");
	LOGD("%s", verbose.c_str());
}

std::ostream &operator<<(std::ostream &stream, const print_errno &p) {
	char buf[256];
	return stream << strerror_r(p.err, buf, sizeof(buf));
}

LogLock::LogLock(ldp_log_level level) {
	locked = tslog::enabled(level);
	if (locked)
		pthread_mutex_lock(&g_mutex);
}

LogLock::~LogLock() {
	if (locked)
		pthread_mutex_unlock(&g_mutex);
}

} // namespace tslog
