#pragma once

namespace type_list {
// this is a list of types.
// It is parameterized by types contained in the list, and provides count field
// which gives number of elements in the list
template <typename ... Types>
struct type_list { constexpr static int count = sizeof...(Types); };

/******************* Helpers for dealing with lists of types ************/
/* Compile-time computing of an index of a type in the type list

   Usage:
   Declare your type list, e.g.
   using my_type_list = type_list<int, char, long, MatchItemOwn>;

   IDX<int, my_type_list>::idx gives 0
   IDX<char, my_type_list>::idx gives 1
   IDX<long, my_type_list>::idx gives 2
   IDX<MatchItemOwn, my_type_list>::idx gives 3

   In some functional language it could be expressed as:

   idx_impl(elem, head::tail, n) =
		if elem == head then
			n
		else
			idx_impl(elem, tail, n+1)

   idx(elem, head::tail) = idx_impl(elem, head::tail, 0)

   Below is IDX and IDX_IMPL which express the same thing in C++.
 */

// idx: look for Elem in TypeList. By default it is a list of MatchItems
template <typename Elem, typename TypeList>
struct IDX { };

namespace _type_list_detail {
// idx_impl: look for Elem in Types, counting how deep we went in n
template <int n, typename Elem, typename ... Types>
struct IDX_IMPL { };

// idx_impl specialization for some type list where head is not the same type as elem
template <int n, typename Elem, typename Head, typename ... Types>
struct IDX_IMPL<n, Elem, Head, Types...> { constexpr static int idx = IDX_IMPL<n+1, Elem, Types...>::idx; };

// idx_impl specialization for some type list where head is the same type as elem
template <int n, typename Elem, typename ... Types>
struct IDX_IMPL<n, Elem, Elem, Types...> { constexpr static int idx = n; };
}

// idx specialization for running idx_impl with all the things extracted
template <typename Elem, template <typename...> typename TypeList, typename ... Types>
struct IDX<Elem, TypeList<Types...>> {
	constexpr static int idx = _type_list_detail::IDX_IMPL<0, Elem, Types...>::idx;
};
} // namespace type_list
