/*
 * Copyright (c) 2015-2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/
/**
 * \file
 * \ingroup Implementation
 */

#ifndef _XML_PARSER_HPP
#define _XML_PARSER_HPP

#include <set>
#include <boost/noncopyable.hpp>
#include "policy.hpp"
#include "global_nodestruct.hpp"
#include <boost/utility/string_ref.hpp>

namespace ldp_xml {
	class StorageBackendXML;
}

namespace ldp_xml_parser
{
	/** Class which have methods parsing policy rules contained in xml files */
	class XmlParser : private boost::noncopyable
	{
	public:
		/** Parses given config file for declared bus type */
		int parsePolicyConfigFile();

		void elementStart(const char *el, const char **attr);

		void elementEnd(const char *el);

		void text(const char *text, int len);

		void setRetCode(int ret);

		explicit XmlParser(ldp_xml::StorageBackendXML &storage, const char *fname);

	private:
		class IncludeItem;

		/** Decides whether a filename describes one of main system.conf/session.conf files */
		bool isMainConfFile(const char *filename);

		/** Parses config file and all files included in it (recursively) */
		void parsePolicyConfigFileInternal(const char *filename);

		/** Parses config file and returns all files included in it */
		void parseXmlFile(const char *filename);

		/** Parses <include> element */
		IncludeItem parseIncludeItem();

		/** Get all the .conf files within included subdirectory, POSIX style as boost::filesystem is not header-only */
		void getIncludedFiles(const std::string& parent_dir, const std::string& incldir, std::vector<std::string>& files);

		ItemBuilder __builder;
		PolicyType policy_type{PolicyType::NONE};
		PolicyTypeValue policy_type_value;

		void parsePolicyTag(const char **attr);
		void parsePolicyAttribute(const char* name, const char* value);

		void parseRule(Decision decision, const char **attr);
		void parseRuleAttribute(const char *name, const char *value);

		typedef enum { IDLE, BUSCONFIG, INCLUDE, INCLUDEDIR, POLICY, RULE } ParseState;

		ParseState state{IDLE};

		bool ignore_always{false};
		bool ignore_missing{false};
		std::string current_text;
		std::string curr_dir;
		int ret_code{-1};
		const char *main_filename{nullptr};

		std::vector<std::string> included_files;
	};

} //namespace

#endif
