/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

#include "kdbus.h"
#include "internal/tslog.hpp"
#include <boost/utility/string_ref.hpp>
#include <cassert>
#include <cerrno>
#include <cstddef>
#include <dbuspolicy1/libdbuspolicy1.h>
#include <fcntl.h>
#include <linux/kdbus.h>
#include <memory>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define CONNECTION_LABEL "libdbuspolicy1-kdbus"

#define KDBUS_POOL_SIZE (1024UL * 1024UL)

#define ALIGN8(l) (((l) + 7) & ~7)
#define ALIGNDN8(l) ((l) & ~7)

#define MAX_DBUS_NAME_LEN 255

constexpr uint64_t KDBUS_ITEM_HEADER_SIZE = offsetof(kdbus_item, data);
constexpr uint64_t KDBUS_ITEM_SIZE(uint64_t s) { return KDBUS_ITEM_HEADER_SIZE + s; }
constexpr uint64_t KDBUS_ITEM_SIZE_ALIGN(uint64_t s) { return ALIGN8(KDBUS_ITEM_SIZE(s)); }

/*
 * kdbus_command:
 * A placeholder for kdbus_cmd_* commands designed to easily allocate them on the stack.
 * It creates a stack buffer of appropriate size, emplaces a command in the buffer
 * and initializes 'size' and 'flags' fields.
 * Template parameters are:
 * T - one of kdbus_cmd_*
 * buffer_additional_size - size of maximum additional space needed for the command;
 *                        by default it is 0, which means that only space needed for the command
 *                        itself will be allocated
 */
template <typename T, size_t buffer_additional_size = 0>
class kdbus_command {
	static constexpr size_t cmd_size = ALIGN8(sizeof(T));
	static constexpr size_t buffer_size = cmd_size + buffer_additional_size;
	alignas(T) uint8_t buffer[buffer_size];
public:
	/* By default commands are initialized with no flags set, and size equal to
	 * the maximum buffer size.
	 * Parameter items_size may be used to assign some smaller value to the size,
	 * e.g. it may be set to 0, if users would like to increment command size by themselves.
	 */
	explicit kdbus_command(uint64_t flags = 0, size_t items_size = buffer_additional_size) {
		auto cmd = get();
		cmd->size = cmd_size + items_size;
		cmd->flags = flags;
	}
	T *get() { return reinterpret_cast<T*>(buffer); }
};

int KdbusConnection::open_bus(const char *path)
{
	fd = open(path, O_RDWR|O_NOCTTY|O_LARGEFILE|O_CLOEXEC);
	return fd;
}

static void set_item_str(kdbus_item *item, const char *str, __u64 str_length, __u64 type)
{
	item->size = KDBUS_ITEM_SIZE(str_length);
	item->type = type;
	memcpy(item->str, str, str_length);
}

int KdbusConnection::hello(uint64_t hello_flags,
						   uint64_t attach_flags_send,
						   uint64_t attach_flags_recv)
{
	kdbus_command<kdbus_cmd_hello, KDBUS_ITEM_SIZE_ALIGN(sizeof(CONNECTION_LABEL))> command(hello_flags);
	auto cmd = command.get();

	cmd->attach_flags_send = attach_flags_send;
	cmd->attach_flags_recv = attach_flags_recv;
	cmd->pool_size = KDBUS_POOL_SIZE;

	set_item_str(cmd->items, CONNECTION_LABEL, sizeof(CONNECTION_LABEL), KDBUS_ITEM_CONN_DESCRIPTION);

	if (ioctl(fd, KDBUS_CMD_HELLO, cmd) < 0)
		return -errno;

	free_by_offset(cmd->offset);

	pool = static_cast<decltype(pool)>(mmap(NULL, KDBUS_POOL_SIZE, PROT_READ, MAP_SHARED, fd, 0));
	if (MAP_FAILED == pool)
		return -errno;

	return 0;
}

int KdbusConnection::connect(const char *path,
							 uint64_t hello_flags,
							 uint64_t attach_flags_send,
							 uint64_t attach_flags_recv)
{
	if (open_bus(path) < 0) {
		LOGE("Error opening bus: %s", path);
		return -1;
	}
	if (hello(hello_flags, attach_flags_send, attach_flags_recv) < 0) {
		LOGE("Error: bus hello failed");
		close(fd);
		return -1;
	}
	return 0;
}

constexpr bool kdbus_is_unique_id(const char* name)
{
	return ':' == name[0];
}

static uint64_t kdbus_unique_id(char const *name)
{
	uint64_t res;
	unsigned i = 2;
	int c;

	/* unique id : ':1.001010'. starting index is 3 */
	while (!(c = name[++i] - '0'));
	res = (unsigned)c;
	while ((c = name[++i] - '0') >= 0)
		res = res*10 + (unsigned)c;
	return res;
}

int KdbusConnection::get_conn_info(const char *bus_name,
								   conn_info_type conn_info_type,
								   KdbusConnectionInfo *info)
{
	kdbus_command<kdbus_cmd_info, KDBUS_ITEM_SIZE_ALIGN(ALIGN8(MAX_DBUS_NAME_LEN+1))> command(0, 0);
	auto cmd = command.get();

	if (conn_info_type == POLICY_CONN_INFO_ALL)
		cmd->attach_flags = (KDBUS_ATTACH_CREDS | KDBUS_ATTACH_NAMES | KDBUS_ATTACH_SECLABEL);
	else
		cmd->attach_flags = KDBUS_ATTACH_NAMES;

	if (kdbus_is_unique_id(bus_name)) {
		cmd->id = kdbus_unique_id(bus_name);
	} else {
		auto l = strlen(bus_name);
		cmd->size += KDBUS_ITEM_SIZE_ALIGN(l+1);
		cmd->id = 0;

		set_item_str(cmd->items, bus_name, l+1, KDBUS_ITEM_NAME);
	}
	if (ioctl(fd, KDBUS_CMD_CONN_INFO, cmd) < 0) {
		if (errno == ENXIO || errno == ESRCH)
			return DBUSPOLICY_RESULT_DEST_NOT_AVAILABLE;

		LOGE("Failed KDBUS_CMD_CONN_INFO: %m\n");
		return DBUSPOLICY_RESULT_KDBUS_ERROR;
	}

	info->setNewData(cmd->offset);

	auto conn_info = reinterpret_cast<kdbus_info *>(pool + cmd->offset);
	uintptr_t items_end = (uintptr_t)conn_info + (size_t)conn_info->size;

	for (auto item = conn_info->items;
			(uintptr_t)item < items_end;
			item = reinterpret_cast<decltype(item)>(ALIGN8((uintptr_t)item + (size_t)item->size))) {
		switch (item->type) {
		case KDBUS_ITEM_CREDS:
			info->setCreds(item->creds.euid, item->creds.egid);
			break;
		case KDBUS_ITEM_SECLABEL:
			info->setLabel(item->str);
			break;
		case KDBUS_ITEM_OWNED_NAME:
			info->addName(item->name.name);
			break;
		}
	}

	return 0;
}

int KdbusConnection::free_by_offset(uint64_t offset)
{
	kdbus_command<kdbus_cmd_free> command;
	auto cmd = command.get();
	cmd->offset = offset;

	if (ioctl(fd, KDBUS_CMD_FREE, cmd) != 0) {
		LOGE("failed KDBUS_CMD_FREE");
		return errno;
	}

	return 0;
}

void KdbusConnectionInfo::release_info()
{
	if (set) {
		my_connection.free_by_offset(free_offset);
		set = false;
	}
}

void KdbusConnectionInfo::setNewData(uint64_t offset)
{
	release_info();

	free_offset = offset;
	set = true;

	k_names.clear();
	_label = nullptr;
	uid_n = UID_INVALID;
	gid_n = GID_INVALID;
}
