/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/

#pragma once

#include <pwd.h>
#include <grp.h>
#include <array>
#include "internal/bus_names_array.hpp"
#include <boost/utility/string_ref.hpp>

class KdbusConnectionInfo;

class KdbusConnection {
	int fd;
	uint8_t *pool;

	int open_bus(const char *path);
	int hello(uint64_t hello_flags,
			  uint64_t attach_flags_send,
			  uint64_t attach_flags_recv);
public:
	KdbusConnection() : fd(0), pool(nullptr) {}
	typedef enum { POLICY_CONN_INFO_ALL, POLICY_CONN_INFO_NAME } conn_info_type;

	void shared_init_fd(int fd_to_set) { fd = fd_to_set; }
	void shared_init_pool(void *pool_to_set) { pool = static_cast<decltype(pool)>(pool_to_set); }
	int connect(const char *path,
				uint64_t hello_flags,
				uint64_t attach_flags_send,
				uint64_t attach_flags_recv);
	int get_conn_info(const char *bus_name,
					  conn_info_type conn_info_type,
					  KdbusConnectionInfo *info);
	int free_by_offset(uint64_t offset);
};

class KdbusConnectionInfo {
	static const uid_t UID_INVALID = -1;
	static const gid_t GID_INVALID = -1;

	char const *_label{nullptr};

	KdbusBusNames k_names;
	uint64_t free_offset{0};
	uid_t uid_n{UID_INVALID};
	gid_t gid_n{GID_INVALID};
	bool set{false};

	KdbusConnection &my_connection;

public:
	explicit KdbusConnectionInfo(KdbusConnection &c) : my_connection{c} {}
	~KdbusConnectionInfo() { release_info(); }

	void release_info();
	void setNewData(uint64_t offset);

	void addName(const char *name) { k_names.push_back(name); }

	void setCreds(uid_t uid, gid_t gid) { uid_n = uid; gid_n = gid; }

	void setLabel(const char *label) { _label = label; }

	int get(const char *destination, KdbusConnection::conn_info_type conn_info_type) {
		return my_connection.get_conn_info(destination, conn_info_type, this);
	}
	const KdbusBusNames &names() const { return k_names; }
	KdbusBusNames &names() { return k_names; }
	uid_t uid() const { return uid_n; }
	gid_t gid() const { return gid_n; }
	const char *label() const { return _label; }
};
