/*
 * Copyright (c) 2015 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/


#include "internal/naive_policy_checker.hpp"
#include "internal/tslog.hpp"
#include "kdbus.h"
#include "libdbuspolicy1-private.h"

#include <assert.h>
#include <boost/utility/string_ref.hpp>
#include <cerrno>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <dbuspolicy1/libdbuspolicy1.h>
#include <fcntl.h>
#include <grp.h>
#include <linux/kdbus.h>
#include <linux/limits.h>
#include <memory>
#include <mutex>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define DBUSPOLICY1_EXPORT __attribute__ ((visibility("default")))

using ldp_xml_parser::DecisionResult;
using ldp_xml_parser::MatchItemSend;
using ldp_xml_parser::MatchItemReceive;
using ldp_xml_parser::MatchItemOwn;
using ldp_xml_parser::NaivePolicyChecker;

typedef enum {
	SYSTEM_BUS = 0,
	SESSION_BUS = 1
} BusType;

/*
 * This is a global structure that gets and keeps credentials from the current process,
 * as they are needed for every policy checking operation.
 */
struct udesc {
	uid_t uid;
	gid_t gid;
	char label[256];

	void init() {
		std::call_once(init_once_done, std::bind(&udesc::init_once, this));
	}

	void change_creds(uid_t uid_, gid_t gid_, const char *label_)
	{
		uid = uid_;
		gid = gid_;
		if (label_)
			strncpy(label, label_, strlen(label_) + 1);
	}

private:
	std::once_flag init_once_done;
	void init_once()
	{
		char buf[1024];
		int attr_fd;
		int r;

		attr_fd = open("/proc/self/attr/current", O_RDONLY);
		if (attr_fd < 0)
			assert(false && "failed open: /proc/self/attr/current");
		r = read(attr_fd, buf, sizeof(buf));
		close(attr_fd);

		if (r < 0 || r >= (long int)sizeof(label)) /* read */
			assert(false && "failed read: /proc/self/attr/current");
		buf[r] = 0;

		uid = getuid();
		gid = getgid();
		snprintf(label, r + 1 /* additional byte for \0 */, "%s", buf);

		tslog::init();
	}
} g_udesc;

extern "C" void __dbuspolicy1_change_creds(uid_t uid, gid_t gid, const char* label)
{
	g_udesc.change_creds(uid, gid, label);
}

/*
 * This class decorates NaivePolicyChecker with initialization mechanisms.
 * Thread-safety is ensured by C++ and static initialization in static member functions.
 */
class Checker {
public:
	Checker(NaivePolicyChecker &checker, const char *primary_file_name, const char *secondary_file_name)
		: _checker{checker}
	{
		std::string primary_file_name_serialized = primary_file_name;
		primary_file_name_serialized.append(".serialized");

		auto ok = _checker.initDb(primary_file_name, primary_file_name_serialized.c_str());
		if (!ok)
			ok = _checker.initDb(secondary_file_name);

		assert(ok && "failed database initialization");

		_checker.updateGroupDb(g_udesc.uid, g_udesc.gid);
	}

	NaivePolicyChecker &checker() { return _checker; }

	// this checks if this process can open a bus owned by bus_owner
	inline bool can_open_owned_by(uid_t bus_owner) {
		return _checker.check(bus_owner, g_udesc.uid, g_udesc.gid, g_udesc.label) == DecisionResult::ALLOW;
	}

	static Checker &system() {
		static Checker checker{policy_checker_system(), system_bus_conf_file_primary(), SYSTEM_BUS_CONF_FILE_SECONDARY};
		return checker;
	}

	static Checker &session() {
		static Checker checker{policy_checker_session(), session_bus_conf_file_primary(), SESSION_BUS_CONF_FILE_SECONDARY};
		return checker;
	}

	static inline Checker &get(BusType bus_type) {
		if (SESSION_BUS == bus_type)
			return session();
		return system();
	}

private:
	NaivePolicyChecker &_checker;
};

/*
 * kconn is a class which binds together a kdbus connection and a corresponding policy checker.
 * It has two flavors:
 * - global objects for system bus and session bus with private connections (static get() method);
 * - objects created on demand when a client shares his kdbus connection with the library (static get_shared() method).
 *
 * Because  the pointer to such object is passed (through void *) to a client, and then clients
 * pass the pointer to dbuspolicy1_* functions, we always need to maintain the same type,
 * regardless of the flavor.
 *
 * Thread-safety is ensured by C++ and static initialization in static member functions.
 */
class kconn {
public:
	KdbusConnection conn;
	NaivePolicyChecker &checker;
	bool can_be_freed;

	kconn(NaivePolicyChecker &ch, const char *resolved_path)
		: checker{ch}, can_be_freed{false}
	{
		if (conn.connect(resolved_path, 0, _KDBUS_ATTACH_ALL, 0) < 0) {
			assert(false && "failed kdbus_connect");
		}
	}

	kconn(NaivePolicyChecker &ch, int fd)
		: checker{ch}, can_be_freed{true}
	{
		conn.shared_init_fd(fd);
	}

	static inline kconn &get(BusType bus_type, const char *resolved_path) {
		if (SYSTEM_BUS == bus_type)
			return system(resolved_path);
		return session(resolved_path);
	}

	static inline kconn *get_shared(BusType bus_type, int fd)
	{
		kconn *result = new kconn(Checker::get(bus_type).checker(), fd);
		if (nullptr == result)
			LOGE("Error: failed to allocate memory for policy configuration");
		return result;
	}

private:
	static kconn &system(const char *resolved_path) {
		static kconn c{Checker::system().checker(), resolved_path};
		return c;
	}
	static kconn &session(const char *resolved_path) {
		static kconn c{Checker::session().checker(), resolved_path};
		return c;
	}

};

/*
 * This class encapsulates functions used to analyze passed path to kdbus bus file.
 * It recognizes system and session paths.
 * If it recognizes path as a session bus, then it extracts owner's uid.
 */
class BusPathResolver {
	std::shared_ptr<const char> resolved_path;
	BusType type{SESSION_BUS};
	uid_t owner{0};
	bool ok{false};

	// We will never get 0 as a session bus owner (0 uses "0-system" bus),
	// so it's OK to assume that bus_owner==0 always means error.
	bool extract_bus_owner() noexcept
	{
		boost::string_ref user_suffix("-user/bus");
		boost::string_ref ps(resolved_path.get());

		if (!ps.ends_with(user_suffix))
			return false;

		ps.remove_suffix(user_suffix.length());

		size_t last_slash = ps.find_last_of('/');
		if (last_slash == boost::string_ref::npos || last_slash == ps.length() - 1)
			return false;

		ps = ps.substr(last_slash+1);

		errno = 0;
		uid_t bus_owner = strtol(ps.data(), NULL, 10);
		if (errno)
			return false;

		owner = bus_owner;
		type = SESSION_BUS;
		return true;
	}
#define STR_SIZE(x) (x), sizeof(x)-1
	static constexpr boost::string_ref KDBUS_PATH_PREFIX{STR_SIZE("/sys/fs/kdbus/")};
	static constexpr boost::string_ref KDBUS_SYSTEM_BUS_PATH{STR_SIZE("/sys/fs/kdbus/0-system/bus")};
#undef STR_SIZE

public:
	BusPathResolver(const char *bus_path)
		: resolved_path(realpath(bus_path, nullptr), [](const char *a) { free(const_cast<char*>(a)); })
	{
		if (!resolved_path)
			return;

		// recognize bus type
		if (KDBUS_SYSTEM_BUS_PATH == resolved_path.get()) {
			type = SYSTEM_BUS;
			ok = true;
		} else if (boost::string_ref(resolved_path.get()).starts_with(KDBUS_PATH_PREFIX)) {
			ok = extract_bus_owner();
		}
	}
	operator bool() { return ok; }
	const char *path() { return resolved_path.get(); }
	BusType bus_type() { return type; }
	uid_t bus_owner() { return owner; }
};

constexpr boost::string_ref BusPathResolver::KDBUS_PATH_PREFIX;
constexpr boost::string_ref BusPathResolver::KDBUS_SYSTEM_BUS_PATH;

DBUSPOLICY1_EXPORT void* dbuspolicy1_init_shared(const char *bus_path, int fd)
{
	assert(bus_path);

	BusPathResolver path_resolver(bus_path);
	if (!path_resolver) {
		LOGE("Error resolving bus path: %s", bus_path);
		return nullptr;
	}

	g_udesc.init();

	auto bus_type = path_resolver.bus_type();

	kconn *result = nullptr;
	if (Checker::get(bus_type).can_open_owned_by(path_resolver.bus_owner())) {
		if (-1 == fd) {
			result = &kconn::get(bus_type, path_resolver.path());
		} else {
			result = kconn::get_shared(bus_type, fd);
		}
	}

	tslog::flush();

	return result;
}

DBUSPOLICY1_EXPORT void* dbuspolicy1_init(const char *bus_path)
{
	return dbuspolicy1_init_shared(bus_path, -1);
}

// helper functions
namespace {
constexpr kconn *KCONN(void *config) { return static_cast<kconn *>(config); }

/*
 * This function prepares contents of info.names() for initialization of MatchItems.
 * First, it tries to get names from kdbus connection.
 * If it fails, names are added from the 'name' parameter, which is treaten
 * as a string containing names separated by single space.
 */
inline int prepareNames(const char *name, KdbusConnectionInfo &info,
		KdbusConnection::conn_info_type info_type) {
	if (name && *name) {
		int r = info.get(name, info_type);
		if (r < 0)
			return r;

		if (info.names().empty())
			info.names().addSpaceSeparatedNames(name);
	}
	return 0;
}

constexpr int decisionToRetCode(DecisionResult decision) {
	return static_cast<int>(decision);
}

constexpr ldp_xml_parser::MessageType toMessageType(int type) {
	return static_cast<ldp_xml_parser::MessageType>(type);
}
} // namespace

DBUSPOLICY1_EXPORT void dbuspolicy1_init_set_pool(void *configuration, void *pool)
{
	KCONN(configuration)->conn.shared_init_pool(pool);
}

DBUSPOLICY1_EXPORT void dbuspolicy1_free(void* configuration)
{
	auto kconn = KCONN(configuration);
	if (kconn->can_be_freed)
		delete kconn;
}

DBUSPOLICY1_EXPORT int dbuspolicy1_check_out(void* configuration,
											 const char *destination,
											 const char *sender,
											 const char *path,
											 const char *interface,
											 const char *member,
											 int         message_type,
											 const char * /*error_name*/,
											 int         /*reply_serial*/,
											 int         /*requested_reply*/)
{
	/* Broadcast signal has NULL destination */
	/* Due to this sender can not check rule correctly */
	if (message_type == DBUSPOLICY_MESSAGE_TYPE_SIGNAL && !destination)
		return 1;

	tslog::LogLock log_lock;
	auto kconn = KCONN(configuration);
	KdbusConnectionInfo destinationInfo(kconn->conn);
	int r;
	/* check can send */
	/* if broadcasting, then pass - null destination */
	r = prepareNames(destination, destinationInfo, KdbusConnection::POLICY_CONN_INFO_ALL);
	if (r < 0)
		return r;

	auto m_type = toMessageType(message_type);

	MatchItemSend sendItem(interface, member, path, m_type, destinationInfo.names());
	auto decision = kconn->checker.check(g_udesc.uid, g_udesc.gid, g_udesc.label, sendItem);

	if (DecisionResult::ALLOW != decision)
		return decisionToRetCode(decision);

	KdbusConnectionInfo senderInfo(kconn->conn);
	/* check can recv */
	/* get sender information from kdbus */
	r = prepareNames(sender, senderInfo, KdbusConnection::POLICY_CONN_INFO_NAME);
	if (r < 0)
		return r;

	MatchItemReceive receiveItem(interface, member, path, m_type, senderInfo.names());
	return decisionToRetCode(kconn->checker.check(destinationInfo.uid(),
												  destinationInfo.gid(),
												  destinationInfo.label(),
												  receiveItem));
}

DBUSPOLICY1_EXPORT int dbuspolicy1_check_in(void* configuration,
											const char *destination,
											const char *sender,
											const char *sender_label,
											uid_t       sender_uid,
											gid_t       sender_gid,
											const char *path,
											const char *interface,
											const char *member,
											int         message_type,
											const char * /*error_name*/,
											int         /*reply_serial*/,
											int         /*requested_reply*/)
{
	tslog::LogLock log_lock;
	auto kconn = KCONN(configuration);
	KdbusConnectionInfo info(kconn->conn);
	int r;
	r = prepareNames(destination, info, KdbusConnection::POLICY_CONN_INFO_NAME);
	if (r < 0)
		return r;

	auto m_type = toMessageType(message_type);

	MatchItemSend sendItem(interface, member, path, m_type, info.names());
	auto decision = kconn->checker.check(sender_uid, sender_gid, sender_label, sendItem);

	if (DecisionResult::ALLOW != decision)
		return decisionToRetCode(decision);

	if (!sender)
		sender = ":";

	/* libdbus, gdbus pass multiple sender as parameter : eg. "name_A name_B name_C". */
	KdbusBusNames names;
	MatchItemReceive receiveItem(interface, member, path, m_type, names.addSpaceSeparatedNames(sender));

	return decisionToRetCode(kconn->checker.check(g_udesc.uid,
												  g_udesc.gid,
												  g_udesc.label,
												  receiveItem));
}

DBUSPOLICY1_EXPORT int dbuspolicy1_can_own(void* configuration, const char* const service)
{
	tslog::LogLock log_lock;
	auto kconn = KCONN(configuration);
	return decisionToRetCode(kconn->checker.check(g_udesc.uid,
												  g_udesc.gid,
												  g_udesc.label,
												  MatchItemOwn(service)));
}
