#include "internal/naive_policy_checker.hpp"
#include "internal/tslog.hpp"
#include <dbuspolicy1/libdbuspolicy1.h>
#include <map>
#include <string>
#include <vector>

using ldp_xml_parser::DecisionResult;

struct AccessTest {
	DecisionResult expected_result;
	uid_t user;
	gid_t group;
	const char* label;
};

const char* DEFAULT_LABEL = "User::Shell";
const char* PRIVILEGED_LABEL = "System::Privileged";

const std::vector<AccessTest> default_allow_tests = {
	{DecisionResult::ALLOW, 0, 0, PRIVILEGED_LABEL},
	{DecisionResult::ALLOW, 2, 20, DEFAULT_LABEL},	//deny user, allow group
	{DecisionResult::ALLOW, 3, 30, DEFAULT_LABEL},	//deny group, allow user
	{DecisionResult::DENY, 4, 30, DEFAULT_LABEL},	//deny group, no user rule
	{DecisionResult::ALLOW, 5, 5, DEFAULT_LABEL},	//no rules
	{DecisionResult::DENY, 6, 6, DEFAULT_LABEL},	//deny user (mandatory)
	{DecisionResult::DENY, 7, 7, DEFAULT_LABEL},	//allow user (def), deny user (mandatory)
	{DecisionResult::ALLOW, 8, 8, DEFAULT_LABEL},	//deny user (def), allow user (mandatory)
	{DecisionResult::ALLOW, 9991, 9991, DEFAULT_LABEL},	//check-user=allow (def)
	{DecisionResult::DENY, 9992, 9992, DEFAULT_LABEL},	//check-user=allow (def), check-group=deny (mandatory)
	{DecisionResult::ALLOW, 9993, 9993, DEFAULT_LABEL},	//check-user=allow (def, mandatory)
	{DecisionResult::ALLOW, 888, 888, PRIVILEGED_LABEL}	//check-user=allow (def) (label)
};

const std::vector<AccessTest> default_deny_tests = {
	{DecisionResult::DENY, 0, 0, PRIVILEGED_LABEL},
	{DecisionResult::ALLOW, 2, 20, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 3, 30, DEFAULT_LABEL},
	{DecisionResult::DENY, 4, 30, DEFAULT_LABEL},
	{DecisionResult::DENY, 5, 5, DEFAULT_LABEL},	// no rules, so denied by default
	{DecisionResult::DENY, 6, 6, DEFAULT_LABEL},
	{DecisionResult::DENY, 7, 7, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 8, 8, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 9991, 9991, DEFAULT_LABEL},
	{DecisionResult::DENY, 9992, 9992, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 9993, 9993, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 888, 888, PRIVILEGED_LABEL}
};

const uid_t bus_owner = 500;

const std::vector<AccessTest> session_tests = {
	{DecisionResult::ALLOW, 500, 500, DEFAULT_LABEL},	// allowed implicitly as the bus owner
	{DecisionResult::DENY, 400, 500, DEFAULT_LABEL},
	{DecisionResult::DENY, 0, 0, PRIVILEGED_LABEL},
	{DecisionResult::DENY, 900, 900, DEFAULT_LABEL},
	{DecisionResult::ALLOW, 600, 600, DEFAULT_LABEL},	// allowed explicitly in policy (user)
	{DecisionResult::ALLOW, 700, 700, DEFAULT_LABEL},	// allowed explicitly in policy (group)
	{DecisionResult::DENY, 800, 700, DEFAULT_LABEL}	// allow group, deny user
};

typedef std::pair<std::string, std::vector<AccessTest>> TestBusSetup;

const std::vector<std::pair<TestBusSetup, TestBusSetup>> access_tests{
	{{"tests/default_deny/system.conf", default_allow_tests}, {"tests/default_deny/session.conf", session_tests}},
	{{"tests/default_deny/system-au.conf", default_allow_tests}, {}},
	{{"tests/default_deny/system-ag.conf", default_allow_tests}, {}},
	{{"tests/default_deny/system-du.conf", default_deny_tests}, {}},
	{{"tests/default_deny/system-dg.conf", default_deny_tests}, {}}
};

std::map<DecisionResult, const char*> DECISIONS {
		{DecisionResult::CYNARA_ERROR, "CYNARA_ERROR"},
		{DecisionResult::DENY, "DENY"},
		{DecisionResult::ALLOW, "ALLOW"}};

void print_test(const struct AccessTest* t, DecisionResult result) {
	printf("uid = %lu, gid = %lu, label = %s, expected = %s, result = %s",
		   (unsigned long)t->user, (unsigned long)t->group, t->label,
		   DECISIONS[t->expected_result], DECISIONS[result]);
}

void run_tests_for_bus(ldp_xml_parser::NaivePolicyChecker &checker, const std::vector<AccessTest>& test_setup, int& i, bool& passed) {
	for (const auto& test : test_setup) {
		checker.updateGroupDb(test.user, test.group);
		auto decision = checker.check(bus_owner, test.user, test.group, test.label);
		if (decision != test.expected_result) {
			printf("[ERROR][%d] access test failed: %s %s ", i, DECISIONS[test.expected_result], DECISIONS[decision]);
			print_test(&test, decision);
			printf("\n");
			passed = false;
		}
		i++;
	}
}

bool run_access_tests() {
	bool passed = true;
	int i = 0;
	for (const auto& test_setup : access_tests) {
		const auto& system_bus_setup = test_setup.first;
		const auto& session_bus_setup = test_setup.second;

		// first initialize both, then perform tests to account for potential (re)initialization bugs
		policy_checker_system().initDb(system_bus_setup.first.c_str());
		if (session_bus_setup.first != "") {
			policy_checker_session().initDb(session_bus_setup.first.c_str());
		}

		run_tests_for_bus(policy_checker_system(), system_bus_setup.second, i, passed);
		run_tests_for_bus(policy_checker_session(), session_bus_setup.second, i, passed);
	}
	return passed;
}

int main() {
	tslog::init();
	if (!run_access_tests())
		return -1;
	return 0;
}
