
#include <iostream>
#include <string>
#include <sys/types.h>
#include <dbuspolicy1/libdbuspolicy1.h>
#include "internal/naive_policy_checker.hpp"
#include "internal/policy.hpp"
#include "internal/tslog.hpp"

using namespace ldp_xml_parser;

enum MessageDirection {
	RECEIVE,
	SEND
};

struct MethodTest {
	DecisionResult expected_result;
	uid_t user;
	gid_t group;
	const char* label;
	const char* name;
	const char* path;
	const char* interface;
	const char* member;
	MessageType type;
	MessageDirection recv_send;
};

/**
 * This test set tests ability to parse xml db
 * and check method call allowance in many use cases
 */

struct MethodTest method_tests[]={
	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test2",  NULL, "org.test.Itest1", "DoIt", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test3",  NULL, "org.test.Itest1", "DoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE},

	{DecisionResult::ALLOW, 5001,  100, "test", "org.test.test3",  NULL, "org.test.Itest1", "DoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE},
	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test2",  NULL, "org.test.Itest1", "DoIt", MessageType::METHOD_CALL, MessageDirection::SEND},

	{DecisionResult::DENY,     0,    0, "test", "org.test.test2",  NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test3",  NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE},

	{DecisionResult::DENY,     0,    0, "test", "org.test.test2",  NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::DENY,  5001,  100, "test", "org.test.test3",  NULL, "org.test.Itest1", "DontDoIt", MessageType::METHOD_CALL, MessageDirection::RECEIVE},

	{DecisionResult::ALLOW,    0,    0, "test", "test.te34.fg4 a.b.c.d.e org.test.test2",  NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::DENY,     0,    0, "test", "test.te34.fg4 a.b.c.d.e",  NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test3",  NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE},

	{DecisionResult::ALLOW,    0,    0, "test", "org.test.test2",  NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::DENY,  5001,  100, "test", "org.test.test3",  NULL, "org.test.Itest1", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE},

	{DecisionResult::DENY,     0,    0, "test", "org.test.test2",  NULL, "org.test.Itest2", "NotKnown", MessageType::METHOD_CALL, MessageDirection::SEND},
	{DecisionResult::ALLOW, 5001,  100, "test", "org.test.test3",  NULL, "org.test.Itest2", "NotKnown", MessageType::METHOD_CALL, MessageDirection::RECEIVE},
};

std::map<DecisionResult, const char*> DECISIONS {
		{DecisionResult::CYNARA_ERROR, "CYNARA_ERROR"},
		{DecisionResult::DENY, "DENY"},
		{DecisionResult::ALLOW, "ALLOW"}};

void methodTest_print(struct MethodTest* t, DecisionResult result) {
	printf("uid = %lu, gid = %lu, label = %s, name = %s, path = %s, interface = %s, member = %s, expected = %s, result = %s  (type=%s)",
		   (unsigned long)t->user, (unsigned long)t->group, t->label,
		   t->name, t->path, t->interface, t->member,
		   DECISIONS[t->expected_result], DECISIONS[result],
		   t->recv_send == MessageDirection::SEND ? "SEND" : "RECEIVE");
}

bool method_test() {
	unsigned  i = 0;
	bool flag = true;
	policy_checker_system().initDb("tests/default_allow/system.conf");
	for (i = 0; i < sizeof(method_tests)/sizeof(struct MethodTest); i++) {
		KdbusBusNames bus_names;
		bus_names.addSpaceSeparatedNames(method_tests[i].name);

		ldp_xml_parser::DecisionResult decision;
		if (method_tests[i].recv_send == MessageDirection::SEND)
		{
			MatchItemSend itemSend(method_tests[i].interface,
								   method_tests[i].member,
								   method_tests[i].path,
								   method_tests[i].type,
								   bus_names);
			decision = policy_checker_system().check(method_tests[i].user,
													 method_tests[i].group,
													 method_tests[i].label,
													 itemSend);
		} else {
			assert(method_tests[i].recv_send == MessageDirection::RECEIVE);
			MatchItemReceive itemReceive(method_tests[i].interface,
										 method_tests[i].member,
										 method_tests[i].path,
										 method_tests[i].type,
										 bus_names);

			decision = policy_checker_system().check(method_tests[i].user,
													 method_tests[i].group,
													 method_tests[i].label,
													 itemReceive);
		}

		if (method_tests[i].expected_result != decision) {
			printf("[ERROR][%d] method test failed: %s %s ",
					i, DECISIONS[method_tests[i].expected_result], DECISIONS[decision]);
			methodTest_print(&method_tests[i], decision);
			printf("\n");
			flag = false;
		}
	}
	return flag;
}

int main() {
	tslog::init();
	if (!method_test())
		return -1;
	return 0;
}
